#!/bin/env python
# -*- coding: utf-8 -*-
##
# ipython_magic.py: Integration into the IPython notebook environment.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

# NB: This should ONLY be imported from an IPython session.

import qsharp as qs
from IPython.display import display
from IPython.core.magic import (register_line_magic, register_cell_magic,
                                register_line_cell_magic, needs_local_scope)


def register_magics():
    @register_cell_magic
    @needs_local_scope
    def qsharp(magic_args, cell, local_ns=None):
        """Compiles a Q# snippet, exposing its operations and functions to
           the current local scope."""
        callables = qs.compile(cell)
        if isinstance(callables, qs.QSharpCallable):
            local_ns[callables._name] = callables
        else:
            for qs_callable in callables:
                local_ns[qs_callable._name] = qs_callable

def register_experimental_magics():
    import qsharp.experimental as exp

    @register_line_magic
    def noise_model(line):
        args = line.split(' ')
        if args[0] == '--set-by-name':
            exp.set_noise_model_by_name(args[1])



#!/bin/env python
# -*- coding: utf-8 -*-
##
# test_loader.py: Tests that the loader correctly handles namespaces and
#     callables.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

## IMPORTS ##

import json
import numpy as np
import os
import pytest
import qsharp
import qsharp.clients.mock
from .utils import set_environment_variables

print ( qsharp.component_versions() )

old_client = qsharp.client

## SETUP ##

def setup_module():
    # Override with the mock client.
    qsharp.client = qsharp.clients.mock.MockClient()
    # Set which operations the mock client will report in response to `%who`.
    qsharp.client.mock_operations = ["A.B.C", "A.B.D", "A.E.F"]

def teardown_module():
    qsharp.client = old_client

## TESTS ##

def test_can_import():
    import A.B
    import A.B

def test_can_import_sub_ns():
    import A

def test_import_dir_is_correct():
    import A
    assert dir(A) == ["B", "E"]

    import A.B
    assert dir(A.B) == ["C", "D"]



# Auto-generated file, do not edit.
##
# version.py: Specifies the version of the qsharp package.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
__version__ = "0.20.2110.171562b1"
is_conda = False
_user_agent_extra = "[0.20.2110.171562b1]"



#!/bin/env python
# -*- coding: utf-8 -*-
##
# serialization.py: Utilities for mapping C# values to and from JSON.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

try:
    import numpy as np
except ImportError:
    np = None
except Exception as ex:
    print("Failed to import numpy. Please check your numpy installation.")
    raise

# Tuples are json encoded differently in C#, this makes sure they are in the right format.
def map_tuples(obj):
    """
    Given a Python object to be serialized, converts any tuples to dictionaries
    of a form expected by the Q# backend.
    """
    if isinstance(obj, tuple):
        result = {
            '@type': 'tuple'
        }
        
        # For tuples of more than 7 items, the .NET type is ValueTuple<T1,T2,T3,T4,T5,T6,T7,TRest>.
        # Items beyond Item7 must be nested inside a key called "Rest".
        max_tuple_length = 7
        for i in range(min(len(obj), max_tuple_length)):
            result[f"Item{i+1}"] = map_tuples(obj[i])
        if len(obj) > max_tuple_length:
            result["Rest"] = map_tuples(obj[max_tuple_length:])
        return result

    elif isinstance(obj, list) or (np and isinstance(obj, np.ndarray)):
        result = []
        for i in obj:
            result.append(map_tuples(i))
        return result

    elif isinstance(obj, dict):
        result = {}
        for i in obj:
            result[i] = map_tuples(obj[i])
        return result

    elif np and isinstance(obj, np.generic):
        return obj.item()

    else:
        return obj

def unmap_tuples(obj):
    """
    Given a Python object deserialized from JSON, converts any dictionaries that
    represent tuples back to Python tuples. Dictionaries are considered to be
    tuples if they either contain a key `@type` with the value `tuple`, or if
    they have a key `item1`.
    """
    if isinstance(obj, dict):
        # Does this dict represent a tuple?
        if obj.get('@type', None) in ('tuple', '@tuple') or 'Item1' in obj:
            values = []
            while True:
                item = f"Item{len(values) + 1}"
                if item in obj:
                    values.append(unmap_tuples(obj[item]))
                else:
                    break
            return tuple(values)
        # Since this is a plain dict, unmap its values and we're good.
        return {
            key: unmap_tuples(value)
            for key, value in obj.items()
        }

    elif isinstance(obj, list):
        return [unmap_tuples(value) for value in obj]

    else:
        return obj



#!/bin/env python
# -*- coding: utf-8 -*-
##
# utils.py: Common functions for use in IQ# Python tests.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

import importlib
import os
import qsharp
import sys

def set_environment_variables():
    '''
    Sets environment variables for test execution and restarts the IQ# kernel.
    '''        
    os.environ["AZURE_QUANTUM_ENV"] = "mock"
    os.environ["IQSHARP_AUTO_LOAD_PACKAGES"] = "$null"
    importlib.reload(qsharp)
    if "qsharp.chemistry" in sys.modules:
        importlib.reload(qsharp.chemistry)



#!/bin/env python
# -*- coding: utf-8 -*-
##
# mock.py: Placeholder client useful in running unit tests and in hosted CI.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import atexit
import json

from typing import List, Dict, Callable, Any, Optional
from distutils.version import LooseVersion
from contextlib import contextmanager

## LOGGING ##

import logging
logger = logging.getLogger(__name__)

## CLASSES ##

class MockClient(object):
    packages: List[str]
    mock_operations: List[str] = ["A.B.C", "A.B.D", "A.E.F"]

    def __init__(self):
        self.packages = []

    ## Server Lifecycle ##

    def start(self):
        logger.debug("MockClient.start called.")
        atexit.register(self.stop)

    def stop(self):
        logger.debug("MockClient.stop called.")

    def is_ready(self):
        logger.debug("MockClient.is_ready called.")
        return True

    def check_status(self):
        logger.debug("MockClient.check_status called.")

    ## Public Interface ##

    @property
    def busy(self) -> bool:
        logger.debug("MockClient.busy accessed.")
        return False

    def compile(self, body):
        logger.debug(f"MockClient.compile called with body:\n{body}")
        return ["Workspace.Snippet.Example"]

    def get_available_operations(self) -> List[str]:
        logger.debug("MockClient.get_available_operations called.")
        return self.mock_operations

    def get_operation_metadata(self, name : str) -> Dict[str, Any]:
        logger.debug(f"MockClient.get_operation_metadata called with name {name}.")
        return {}

    def get_workspace_operations(self) -> List[str]:
        logger.debug("MockClient.get_workspace_operations called.")
        return []

    def reload(self) -> None:
        logger.debug("MockClient.reload called.")
        return None

    def add_package(self, name : str) -> None:
        logger.debug(f"MockClient.add_package called with name {name}.")
        return self.packages.append(name)

    def get_packages(self) -> List[str]:
        logger.debug("MockClient.get_packages called.")
        return self.packages

    def simulate(self, op, **params) -> Any:
        logger.debug(f"MockClient.simulate called with operation {op} and params:\n{params}")
        return ()

    def toffoli_simulate(self, op, **params) -> Any:
        logger.debug(f"MockClient.toffoli_simulate called with operation {op} and params:\n{params}")
        return ()

    def estimate(self, op, **params) -> Dict[str, int]:
        logger.debug(f"MockClient.estimate called with operation {op} and params:\n{params}")
        return {
            "Depth": 13,
            "Width": 15
        }

    def component_versions(self, **kwargs) -> Dict[str, LooseVersion]:
        """
        Returns a dictionary from components of the IQ# kernel to their
        versions.
        """
        logger.debug(f"MockClient.component_versions called with keyword arguments:\n{kwargs}")
        return {}

    @contextmanager
    def capture_diagnostics(self, passthrough: bool) -> List[Any]:
        data = []
        yield data



#!/bin/env python
# -*- coding: utf-8 -*-
##
# test_serialization.py: Checks correctness of JSON serialization.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

## IMPORTS ##

import unittest
import json
import numpy as np
import pytest
from qsharp.serialization import map_tuples, unmap_tuples
from .utils import set_environment_variables

## SETUP ##

@pytest.fixture(scope="session", autouse=True)
def session_setup():
    set_environment_variables()

## TESTS ##

class TestSerialization(unittest.TestCase):
    def test_map_shallow_tuple(self):
        self.assertEqual(
            map_tuples((42, 'foo')),
            {'@type': 'tuple', 'Item1': 42, 'Item2': 'foo'}
        )

    def test_map_deep_tuple(self):
        actual = {
            'foo': [1, 3.14, (42, 'baz')],
            'bar': {'a': ('a', 'a'), 'b': ()}
        }
        expected = {
            'foo': [1, 3.14, {'@type': 'tuple', 'Item1': 42, 'Item2': 'baz'}],
            'bar': {
                'a': {'@type': 'tuple', 'Item1': 'a', 'Item2': 'a'},
                'b': {'@type': 'tuple'}
            }
        }
        self.assertEqual(
            map_tuples(actual), expected
        )


    def test_map_numpy_types(self):
        self.assertEqual(
            map_tuples(np.array([1, 2, 3], dtype=np.int32)),
            [1, 2, 3]
        )
        self.assertEqual(
            map_tuples(np.array([1., 2., 3.], dtype=np.single)),
            [1., 2., 3.]
        )


    def test_map_ndarray(self):
        self.assertEqual(
            map_tuples(np.array([1, 2, 3])),
            [1, 2, 3]
        )
        
        tuples = [(0, 'Zero'), (1, 'One')]
        tuples_array = np.empty(len(tuples), dtype=object)
        tuples_array[:] = tuples
        self.assertEqual(
            map_tuples(tuples_array),
            [
                {'@type': 'tuple', 'Item1': 0, 'Item2': 'Zero'},
                {'@type': 'tuple', 'Item1': 1, 'Item2': 'One'}
            ]
        )

    def test_roundtrip_shallow_tuple(self):
        actual = ('a', 3.14, False)
        self.assertEqual(
            unmap_tuples(map_tuples(actual)), actual
        )

    def test_roundtrip_dict(self):
        actual = {'a': 'b', 'c': ('d', 'e')}
        self.assertEqual(
            unmap_tuples(map_tuples(actual)), actual
        )

    def test_roundtrip_deep_tuple(self):
        actual = ('a', ('b', 'c'))
        self.assertEqual(
            unmap_tuples(map_tuples(actual)), actual
        )

    def test_roundtrip_very_deep_tuple(self):
        actual = {
            'a': {
                'b': (
                    {
                        'c': ('d', ['e', ('f', 'g', 12, False)])
                    },
                    ['h', {'g': ('i', 'j')}]
                )
            }
        }
        self.assertEqual(
            unmap_tuples(map_tuples(actual)), actual
        )

if __name__ == "__main__":
    unittest.main()



#!/bin/env python
# -*- coding: utf-8 -*-
##
# config.py: Provides access to IQ# configuration settings.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import Any

## CLASSES ##

class Config(object):
    """
    Provides dictionary-like access to IQ# configuration settings.
    """

    def __init__(self, client):
        self._client = client

    def __getitem__(self, name : str) -> Any:
        """
        Returns the value of the specified IQ# configuration option.
        Options can be set by calls to qsharp.set_config() or by loading
        a file previously created by qsharp.save_config().
        See https://docs.microsoft.com/qsharp/api/iqsharp-magic/config for the list
        of supported IQ# configuration setting names and values.
        """
        return self._client.get_config()[name]

    def __setitem__(self, name : str, value : object) -> None:
        """
        Sets a specified IQ# configuration option with the specified value.
        See https://docs.microsoft.com/qsharp/api/iqsharp-magic/config for the list
        of supported IQ# configuration setting names and values.
        """
        self._client.set_config(name, value)

    def save(self) -> None:
        """
        Saves all current IQ# configuration options to a file named .iqsharp-config.json
        in the current working directory. This file is automatically loaded
        by the IQ# kernel at initialization time.
        See https://docs.microsoft.com/qsharp/api/iqsharp-magic/config for the list
        of supported IQ# configuration setting names and values.
        """
        self._client.save_config()



#!/bin/env python
# -*- coding: utf-8 -*-
##
# iqsharp.py: Client for the IQ# Jupyter kernel.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

## IMPORTS ##

from contextlib import contextmanager
import subprocess
import time
import http.client
import atexit
import json
import sys
import urllib.parse
import os
import jupyter_client

from functools import partial
from io import StringIO
from collections import defaultdict
from typing import List, Dict, Callable, Any, Optional
from pathlib import Path
from distutils.version import LooseVersion

from qsharp.serialization import map_tuples, unmap_tuples

try:
    from IPython.display import display
    display_raw = partial(display, raw=True)
except:
    def display_raw(content):
        pass

## VERSION REPORTING ##

try:
    from qsharp.version import _user_agent_extra
except ImportError:
    _user_agent_extra = ""

## LOGGING ##

import logging
logger = logging.getLogger(__name__)

## CLASSES ##

class IQSharpError(RuntimeError):
    """
    Represents a Q# error passed by the IQ# kernel to the Python host.
    """
    def __init__(self, iqsharp_errors : List[str]):
        self.iqsharp_errors = iqsharp_errors
        error_msg = StringIO()
        error_msg.write("The Q# kernel raised the following errors:\n")
        error_msg.writelines([
            "    " + msg for msg in iqsharp_errors
        ])
        super().__init__(error_msg.getvalue())

class AlreadyExecutingError(IOError):
    """
    Raised when the IQ# client is already executing a command and cannot safely
    process an additional command.
    """
    pass

class IQSharpClient(object):
    kernel_manager = None
    kernel_client = None
    _busy : bool = False

class IQSharpClient(object):
    kernel_manager = None
    kernel_client = None
    _busy : bool = False

    display_data_callback: Optional[Callable[[Any], bool]] = None

    def __init__(self, kernel_name: str = 'iqsharp'):
        self.kernel_manager = jupyter_client.KernelManager(kernel_name=kernel_name)

    ## Server Lifecycle ##

    def start(self):
        logger.info("Starting IQ# kernel...")
        self.kernel_manager.start_kernel(extra_arguments=["--user-agent", f"qsharp.py{_user_agent_extra}"])
        self.kernel_client = self.kernel_manager.client()
        atexit.register(self.stop)

    def stop(self):
        # Don't use logger here. If we're running inside pytest, the handle to the
        # log output file may have already been closed.
        try:
            self.kernel_manager.shutdown_kernel()
        except:
            pass

    def is_ready(self):
        try:
            result = self.component_versions(timeout=6)
            logger.info(f"Q# version\n{result}")
        except Exception as ex:
            logger.info('Exception while checking if IQ# is ready.', exc_info=ex)
            return
        return True

    def check_status(self):
        if not self.kernel_manager.is_alive():
            logger.debug("IQ# kernel is not running. Restarting.")
            self.start()

    ## Public Interface ##

    @property
    def busy(self) -> bool:
        return self._busy

    def compile(self, body):
        return self._execute(body)

    def get_available_operations(self) -> List[str]:
        return self._execute('%who', raise_on_stderr=False)

    def get_operation_metadata(self, name : str) -> Dict[str, Any]:
        return self._execute(f"?{name}")

    def get_workspace_operations(self) -> List[str]:
        return self._execute("%workspace")

    def reload(self) -> None:
        return self._execute(f"%workspace reload", raise_on_stderr=True)

    def get_config(self) -> Dict[str, object]:
        raw = self._execute(f"%config", raise_on_stderr=True)
        config_settings = {}
        for row in raw["rows"]:
            config_settings[row['Key']] = row['Value']
        return config_settings

    def set_config(self, name : str, value : object) -> None:
        from numbers import Number
        if isinstance(value, bool):
            self._execute(f"%config {name}={'true' if value else 'false'}", raise_on_stderr=True)
        elif isinstance(value, Number):
            self._execute(f"%config {name}={value}", raise_on_stderr=True)
        else:
            self._execute(f"%config {name}='{value}'", raise_on_stderr=True)

    def save_config(self) -> None:
        self._execute(f"%config --save", raise_on_stderr=True)

    def add_package(self, name : str) -> None:
        return self._execute(f"%package {name}", raise_on_stderr=True)

    def get_packages(self) -> List[str]:
        return self._execute("%package", raise_on_stderr=False)

    def add_project(self, path : str) -> None:
        return self._execute(f"%project {path}", raise_on_stderr=True)

    def get_projects(self) -> List[str]:
        return self._execute("%project", raise_on_stderr=False)

    def simulate(self, op, **kwargs) -> Any:
        return self._execute_callable_magic('simulate', op, **kwargs)

    def toffoli_simulate(self, op, **kwargs) -> Any:
        return self._execute_callable_magic('toffoli', op, **kwargs)

    def estimate(self, op, **kwargs) -> Dict[str, int]:
        raw_counts = self._execute_callable_magic('estimate', op, **kwargs)
        # Note that raw_counts will have the form:
        # [
        #     {"Metric": "<name>", "Sum": "<value>"},
        #     ...
        # ]
        # We thus need to convert it into a dict. As we do so, we convert counts
        # to ints, since they get turned to floats by JSON serialization.
        counts = {}
        for row in raw_counts:
            counts[row["Metric"]] = int(row["Sum"])
        return counts

    def trace(self, op, **kwargs) -> Any:
        return self._execute_callable_magic('trace', op, _quiet_ = True, **kwargs)

    def component_versions(self, **kwargs) -> Dict[str, LooseVersion]:
        """
        Returns a dictionary from components of the IQ# kernel to their
        versions.
        """
        versions = {}
        def capture(msg):
            # We expect a display_data with the version table.
            if msg["msg_type"] == "display_data":
                data = unmap_tuples(json.loads(self._get_qsharp_data(msg["content"])))
                for component, version in data["rows"]:
                    versions[component] = LooseVersion(version)
        self._execute("%version", display_data_handler=capture, _quiet_=True, **kwargs)
        return versions

    @contextmanager
    def capture_diagnostics(self, passthrough: bool) -> List[Any]:
        captured_data = []
        def callback(msg):
            msg_data = (
                # Check both the old and new MIME types used by the IQ#
                # kernel.
                json.loads(msg['content']['data'].get('application/json', "null")) or
                json.loads(msg['content']['data'].get('application/x-qsharp-data', "null"))
            )
            if msg_data is not None:
                captured_data.append(msg_data)
                return passthrough
            else:
                # No JSON found found, so just fall back.
                return True

        old_callback = self.display_data_callback
        self.display_data_callback = callback
        try:
            yield captured_data
        finally:
            self.display_data_callback = old_callback

    ## Experimental Methods ##
    # These methods expose experimental functionality that may be removed without
    # warning. To communicate to users that these are not reliable, we mark
    # these methods as private, and will re-export them in the
    # qsharp.experimental submodule.

    def _simulate_noise(self, op, **kwargs) -> Any:
        return self._execute_callable_magic('experimental.simulate_noise', op, **kwargs)

    def _get_noise_model(self) -> str:
        return self._execute(f'%experimental.noise_model')

    def _get_noise_model_by_name(self, name : str) -> None:
        return self._execute(f'%experimental.noise_model --get-by-name {name}')

    def _set_noise_model(self, json_data : str) -> None:
        # We assume json_data is already serialized, so that we skip the support
        # provided by _execute_magic and call directly.
        return self._execute(f'%experimental.noise_model {json_data}')

    def _set_noise_model_by_name(self, name : str) -> None:
        return self._execute(f'%experimental.noise_model --load-by-name {name}')


    ## Internal-Use Methods ##

    @staticmethod
    def _get_qsharp_data(message_content):
        if "application/x-qsharp-data" in message_content["data"]:
            # Current versions of IQ# use application/x-qsharp-data
            # for the JSON-encoded data in the execution result.
            return message_content["data"]["application/x-qsharp-data"]
        if "application/json" in message_content["data"]:
            # For back-compat with older versions of IQ# <= 0.17.2105.144881
            # that used the application/json MIME type for the JSON data.
            return message_content["data"]["application/json"]
        return None

    def _execute_magic(self, magic : str, raise_on_stderr : bool = False, _quiet_ : bool = False, **kwargs) -> Any:
        return self._execute(
            f'%{magic} {json.dumps(map_tuples(kwargs))}',
            raise_on_stderr=raise_on_stderr, _quiet_=_quiet_
        )

    def _execute_callable_magic(self, magic : str, op,
            raise_on_stderr : bool = False,
            _quiet_ : bool = False,
            **kwargs
    ) -> Any:
        return self._execute_magic(
            f"{magic} {op._name}",
            raise_on_stderr=raise_on_stderr,
            _quiet_=_quiet_,
            **kwargs
        )

    def _handle_message(self, msg, handlers=None, error_callback=None, fallback_hook=None):
        if handlers is None:
            handlers = {}
        if fallback_hook is None:
            fallback_hook = self.kernel_client._output_hook_default
        msg_type = msg['msg_type']
        if msg_type in handlers:
            handlers[msg_type](msg)
        else:
            if error_callback is not None and msg['msg_type'] == 'stream' and msg['content']['name'] == 'stderr':
                error_callback(msg['content']['text'])
            else:
                fallback_hook(msg)

    def _execute(self, input, return_full_result=False, raise_on_stderr : bool = False, output_hook=None, display_data_handler=None, _quiet_ : bool = False, **kwargs):

        logger.debug(f"sending:\n{input}")

        # make sure the server is still running:
        try:
            self.check_status()
        except:
            raise IQSharpError(["IQ# is not running."])

        results = []
        errors = []

        def log_error(msg):
            errors.append(msg)

        # Set up handlers for various kinds of messages, making sure to
        # fallback through to output_hook as appropriate, so that the IPython
        # package can send display data through to Jupyter clients.
        handlers = {
            'execute_result': (lambda msg: results.append(msg)),
            'render_execution_path':  (lambda msg: results.append(msg)),
            'display_data': display_data_handler if display_data_handler is not None else lambda msg: ...
        }

        # Pass display data through to IPython if we're not in quiet mode.
        if not _quiet_:
            handlers['display_data'] = (
                lambda msg: display_raw(msg['content']['data'])
            )

        # Finish setting up handlers by allowing the display_data_callback
        # to intercept display data first, only sending messages through to
        # other handlers if it returns True.
        if self.display_data_callback is not None:
            inner_handler = handlers['display_data']

            def filter_display_data(msg):
                if self.display_data_callback(msg):
                    return inner_handler(msg)

            handlers['display_data'] = filter_display_data

        _output_hook = partial(
            self._handle_message,
            error_callback=log_error if raise_on_stderr else None,
            fallback_hook=output_hook,
            handlers=handlers
        )

        try:
            if self._busy:
                # Trying to execute while already executing can corrupt the
                # ordering of messages internally to ZeroMQ
                # (see https://github.com/Microsoft/QuantumLibraries/issues/69),
                # so we need to throw early rather than letting the problem
                # propagate to a Jupyter protocol error.
                raise AlreadyExecutingError("Cannot execute through the IQ# client while another execution is completing.")
            self._busy = True
            reply = self.kernel_client.execute_interactive(input, output_hook=_output_hook, **kwargs)
        finally:
            self._busy = False

        logger.debug(f"received:\n{reply}")

        # There should be either zero or one execute_result messages.
        if errors:
            raise IQSharpError(errors)
        if results:
            assert len(results) == 1
            content = results[0]['content']
            if 'executionPath' in content:
                obj = content['executionPath']
            else:
                qsharp_data = self._get_qsharp_data(content)
                if qsharp_data:
                    obj = unmap_tuples(json.loads(qsharp_data))
                else:
                    obj = None
            return (obj, content) if return_full_result else obj
        else:
            return None



#!/bin/env python
# -*- coding: utf-8 -*-
##
# packages.py: Abstraction to represent the list of packages.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from distutils.version import LooseVersion
from typing import Iterable, Tuple

## LOGGING ##

import logging
logger = logging.getLogger(__name__)


## CLASSES ##

class Packages(object):
    """
    Represents the list of packages loaded into the current Q# session, and
    allows for adding new packages from NuGet.org or any other configured feeds.
    """

    def __init__(self, client):
        self._client = client

    def __iter__(self) -> Iterable[Tuple[str, LooseVersion]]:
        for pkg_spec in self._client.get_packages():
            name, version = pkg_spec.split("::", 1)
            yield name, LooseVersion(version)

    def __repr__(self) -> str:
        return repr(list(self))
    def __str__(self) -> str:
        return str(list(self))

    def add(self, package_name : str) -> None:
        """
        Adds a NuGet package with the given package name to the current Q#
        session, downloading the package from NuGet.org or any other configured
        feeds as necessary.
        """
        logger.info(f"Loading package: {package_name}")
        pkgs = self._client.add_package(package_name)
        logger.info("Loading complete: " + ';'.join(str(e) for e in pkgs))



#!/bin/env python
# -*- coding: utf-8 -*-
##
# __init__.py: Logic for launching and configuring Q# clients.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##


import os
import sys
import time
import logging
import jupyter_client
from distutils.util import strtobool

class IQSharpNotInstalledError(Exception):
    pass

class IQSharpNotAvailableError(Exception):
    pass

def _start_client():
    logger = logging.getLogger(__name__)

    client_name =  os.getenv("QSHARP_PY_CLIENT", "iqsharp")

    if client_name == "iqsharp":
        # Allow users to override what kernel is used, making it easier to
        # test kernels side-by-side.
        kernel_name =  os.getenv("QSHARP_PY_IQSHARP_KERNEL_NAME", "iqsharp")
        import qsharp.clients.iqsharp
        client = qsharp.clients.iqsharp.IQSharpClient(kernel_name=kernel_name)
    elif client_name == "mock":
        import qsharp.clients.mock
        client = qsharp.clients.mock.MockClient()

    try:
        client.start()
    except jupyter_client.kernelspec.NoSuchKernel as ex:
        message = "IQ# is not installed." + \
            "\nPlease follow the instructions at https://aka.ms/qdk-install/python."
        print(message)
        raise IQSharpNotInstalledError(message)

    # Check if the server is up and running:
    server_ready = False
    for idx_attempt in range(20):
        try:
            server_ready = client.is_ready()
            if server_ready:
                break
            if idx_attempt == 0:
                print("Preparing Q# environment...")
            else:
                print(".", end='', flush=True)
            time.sleep(1)
        except Exception as ex:
            logger.debug('Exception while checking Q# environment.', exc_info=ex)
            print("!", end='', flush=True)
            time.sleep(1)
    if not server_ready:
        message = "Q# environment was not available in allocated time." + \
            "\nPlease check the instructions at https://aka.ms/qdk-install/python."
        print(message)
        raise IQSharpNotAvailableError(message)

    return client



#!/bin/env python
# -*- coding: utf-8 -*-
##
# tomography.py: Single qubit tomography of Q# operations.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

## IMPORTS ##

import qsharp
import numpy as np

try:
    import qinfer as qi
except ImportError:
    qi = None

try:
    import qutip as qt
except ImportError:
    qt = None

def projector(P):
    if qt is None:
        raise ImportError("This function requires QuTiP.")

    return (qt.qeye(2) + P) / 2.0

def single_qubit_process_tomography(
    operation,
    n_measurements=2000,
    n_particles=4000
):
    """
    :param operation: A Q# operation of type ((Pauli, Pauli) => Result) whose
        inputs are named `prep` and `meas` (respectively), representing
        a state preparation, evolution, and measurement.
    """
    if qt is None:
        raise ImportError("This function requires QuTiP.")
    if qi is None:
        raise ImportError("This function requires QInfer.")

    print("Preparing tomography model...")
    state_basis = qi.tomography.pauli_basis(1)
    prior = qi.tomography.BCSZChoiDistribution(state_basis)
    model = qi.tomography.TomographyModel(prior.basis)

    updater = qi.SMCUpdater(model, n_particles, prior)

    print("Performing tomography...")
    for idx_experiment in range(n_measurements):
        prep = qsharp.Pauli.sample()
        meas = qsharp.Pauli.sample()

        # Convert into a QuTiP object by using the standard transformation
        # between state and process tomography.
        qobj = 2.0 * qt.tensor(
            projector(prep.as_qobj()).trans(), projector(meas.as_qobj())
        )
        expparams = np.array(
            [(model.basis.state_to_modelparams(qobj),)],
            dtype=model.expparams_dtype
        )

        datum = 1 - operation.simulate(prep=prep, meas=meas)

        updater.update(datum, expparams)

    return {
        # We multiply by 2 to turn into a Choi–Jamiłkowski operator instead
        # of a Choi–Jamiłkowski state.
        'est_channel': 2.0 * model.basis.modelparams_to_state(updater.est_mean()),
        # Returning the updater allows for exploring properties not extracted
        # elsewhere.
        'posterior': updater
    }



#!/bin/env python
# -*- coding: utf-8 -*-
##
# test_azure.py: Tests Azure Quantum functionality against a mock workspace.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

## IMPORTS ##

import pytest
import qsharp
from qsharp.azure import AzureError, AzureJob, AzureTarget
from .utils import set_environment_variables

## SETUP ##

@pytest.fixture(scope="session", autouse=True)
def session_setup():
    set_environment_variables()

## TESTS ##

def test_empty_workspace():
    """
    Tests behavior of a mock workspace with no providers.
    """
    with pytest.raises(AzureError) as exception_info:
        qsharp.azure.target()
    assert exception_info.value.error_name == "NotConnected"

    targets = qsharp.azure.connect(
        storage="test",
        subscription="test",
        resourceGroup="test",
        workspace="test",
        location="test"
    )
    assert targets == []

    with pytest.raises(AzureError) as exception_info:
        qsharp.azure.target("invalid.target")
    assert exception_info.value.error_name == "InvalidTarget"

    jobs = qsharp.azure.jobs()
    assert jobs == []

def test_workspace_create_with_no_location():
    """
    Tests behavior of a mock workspace with no location.
    """
    with pytest.raises(AzureError) as exception_info:
        qsharp.azure.connect(
            storage="test",
            subscription="test",
            resourceGroup="test",
            workspace="test"
        )
    assert exception_info.value.error_name == "NoWorkspaceLocation"

def test_workspace_create_with_parameters():
    """
    Tests behavior of a mock workspace with providers, using parameters to connect.
    """
    targets = qsharp.azure.connect(
        storage="test",
        subscription="test",
        resourceGroup="test",
        workspace="WorkspaceNameWithMockProviders",
        location="test"
    )
    assert isinstance(targets, list)
    assert len(targets) > 0

    _test_workspace_with_providers_after_connection()

def test_workspace_create_with_resource_id():
    """
    Tests behavior of a mock workspace with providers, using resource ID to connect.
    Also verifies case-insensitivity of resource ID parsing.
    """
    subscriptionId = "f846b2bd-d0e2-4a1d-8141-4c6944a9d387"
    resourceGroupName = "test"
    workspaceName = "WorkspaceNameWithMockProviders"
    location = "test"
    targets = qsharp.azure.connect(
        resourceId=f"/subscriptions/{subscriptionId}/RESOurceGroups/{resourceGroupName}/providers/Microsoft.Quantum/Workspaces/{workspaceName}",
        location=location)
    assert isinstance(targets, list)
    assert len(targets) > 0

    _test_workspace_with_providers_after_connection()
    _test_workspace_job_execution()

def test_workspace_create_with_resource_id_and_storage():
    """
    Tests behavior of a mock workspace with providers, using resource ID and storage connection string to connect.
    """
    subscriptionId = "f846b2bd-d0e2-4a1d-8141-4c6944a9d387"
    resourceGroupName = "test"
    workspaceName = "WorkspaceNameWithMockProviders"
    storageAccountConnectionString = "test"
    location = "test"
    targets = qsharp.azure.connect(
        resourceId=f"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/Workspaces/{workspaceName}",
        storage=storageAccountConnectionString,
        location=location)
    assert isinstance(targets, list)
    assert len(targets) > 0

    _test_workspace_with_providers_after_connection()

def _test_workspace_with_providers_after_connection():
    with pytest.raises(AzureError) as exception_info:
        qsharp.azure.target()
    assert exception_info.value.error_name == "NoTarget"

    targets = qsharp.azure.connect()
    for target in targets:
        active_target = qsharp.azure.target(target.id)
        assert isinstance(active_target, AzureTarget)
        assert active_target == target

    # Submit a snippet operation without parameters
    op = qsharp.compile("""
        operation HelloQ() : Result
        {
            Message($"Hello from quantum world!"); 
            return Zero;
        }
    """)

    job = qsharp.azure.submit(op)
    assert isinstance(job, AzureJob)

    retrieved_job = qsharp.azure.status(job.id)
    assert isinstance(retrieved_job, AzureJob)
    assert job.id == retrieved_job.id

def _test_workspace_job_execution():
    # Execute a workspace operation with parameters
    op = qsharp.QSharpCallable("Microsoft.Quantum.SanityTests.HelloAgain", None)

    with pytest.raises(AzureError) as exception_info:
        qsharp.azure.execute(op)
    assert exception_info.value.error_name == "JobSubmissionFailed"

    histogram = qsharp.azure.execute(op, count=3, name="test", timeout=3, poll=2)
    assert isinstance(histogram, dict)

    retrieved_histogram = qsharp.azure.output()
    assert isinstance(retrieved_histogram, dict)
    assert histogram == retrieved_histogram

    # Check that the submitted job exists in the workspace
    jobs = qsharp.azure.jobs()
    assert isinstance(jobs, list)
    assert len(jobs) == 1

    # Check that job filtering works
    jobs = qsharp.azure.jobs(jobs[0].id)
    assert isinstance(jobs, list)
    assert len(jobs) == 1
    
    # Check that job count works
    jobs = qsharp.azure.jobs(count=1)
    assert isinstance(jobs, list)
    assert len(jobs) == 1

    jobs = qsharp.azure.jobs("invalid", count=10)
    assert isinstance(jobs, list)
    assert len(jobs) == 0



#!/bin/env python
# -*- coding: utf-8 -*-
##
# projects.py: Abstraction to represent the list of projects.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import Iterable, Tuple

## LOGGING ##

import logging
logger = logging.getLogger(__name__)


## CLASSES ##

class Projects(object):
    """
    Represents the list of projects loaded into the current Q# session, and
    allows for adding references to additional .csproj files.
    """

    def __init__(self, client):
        self._client = client

    def __iter__(self) -> Iterable[str]:
        for project in self._client.get_projects():
            yield project

    def __repr__(self) -> str:
        return repr(list(self))
    def __str__(self) -> str:
        return str(list(self))

    def add(self, project_path : str) -> None:
        """
        Adds a reference to the given Q# project to be loaded
        into the current IQ# session.

        :param project_path: Path to the .csproj to be added. May be an absolute
            path or a path relative to the current workspace root folder.
        """
        logger.info(f"Loading project: {project_path}")
        loaded_projects = self._client.add_project(project_path)
        logger.info("Loading complete: " + ';'.join(str(p) for p in loaded_projects))



#!/bin/env python
# -*- coding: utf-8 -*-
##
# utils.py: Utilities internal to the qsharp package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import logging
logger = logging.getLogger(__name__)
from typing import Callable

## INTERNAL FUNCTIONS ##

def log_messages(data, action : Callable[[str], None] = logger.error):
    msgs = data['messages']
    for msg in msgs:
        action(msg)



#!/bin/env python
# -*- coding: utf-8 -*-
##
# azure.py: enables using Q# quantum execution on Azure Quantum from Python.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

## IMPORTS ##

import qsharp
import json
from typing import List, Dict, Callable, Any, Union
from enum import Enum

## LOGGING ##

import logging
logger = logging.getLogger(__name__)

## EXPORTS ##

__all__ = [
    'connect',
    'target',
    'submit',
    'execute',
    'status',
    'output',
    'jobs',
    'AzureTarget',
    'AzureJob',
    'AzureError'
]

## CLASSES ##

class AzureTarget(object):
    """
    Represents an instance of an Azure Quantum execution target for Q# job submission.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.id = data["id"]
        self.current_availability = data["current_availability"]
        self.average_queue_time = data["average_queue_time"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureTarget):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

class AzureJob(object):
    """
    Represents an instance of an Azure Quantum job.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.id = data["id"]
        self.name = data["name"]
        self.status = data["status"]
        self.uri = data["uri"]
        self.provider = data["provider"]
        self.target = data["target"]
        self.creation_time = data["creation_time"]
        self.begin_execution_time = data["begin_execution_time"]
        self.end_execution_time = data["end_execution_time"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureJob):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

class AzureError(Exception):
    """
    Contains error information resulting from an attempt to interact with Azure.
    """
    def __init__(self, data: Dict):
        self.__dict__ = data
        self.error_code = data["error_code"]
        self.error_name = data["error_name"]
        self.error_description = data["error_description"]

    def __repr__(self) -> str:
        return self.__dict__.__repr__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, AzureError):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.__dict__ == other.__dict__

## FUNCTIONS ##

def connect(**params) -> List[AzureTarget]:
    """
    Connects to an Azure Quantum workspace or displays current connection status.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.connect for more details.
    """
    result = qsharp.client._execute_magic(f"azure.connect", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return [AzureTarget(target) for target in result]

def target(name : str = '', **params) -> AzureTarget:
    """
    Sets or displays the active execution target for Q# job submission in an Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.target for more details.
    """
    result = qsharp.client._execute_magic(f"azure.target {name}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureTarget(result)

def submit(op : qsharp.QSharpCallable, **params) -> AzureJob:
    """
    Submits a job to an Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.submit for more details.
    """
    result = qsharp.client._execute_callable_magic("azure.submit", op, raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureJob(result)

def execute(op : qsharp.QSharpCallable, **params) -> Dict:
    """
    Submits a job to an Azure Quantum workspace and waits for completion.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.execute for more details.
    """
    result = qsharp.client._execute_callable_magic("azure.execute", op, raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return result

def status(jobId : str = '', **params) -> AzureJob:
    """
    Displays status for a job in the current Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.status for more details.
    """
    result = qsharp.client._execute_magic(f"azure.status {jobId}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return AzureJob(result)

def output(jobId : str = '', **params) -> Dict:
    """
    Displays results for a job in the current Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.output for more details.
    """
    result = qsharp.client._execute_magic(f"azure.output {jobId}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return result

def jobs(filter : str = '', count : int = 30, **params) -> List[AzureJob]:
    """
    Displays a list of jobs in the current Azure Quantum workspace.
    See https://docs.microsoft.com/qsharp/api/iqsharp-magic/azure.jobs for more details.
    """
    result = qsharp.client._execute_magic(f"azure.jobs \"{filter}\" count={count}", raise_on_stderr=False, **params)
    if "error_code" in result: raise AzureError(result)
    return [AzureJob(job) for job in result]



# -*- coding: utf-8 -*-
##
# qobj.py: Functions for converting to and from QuTiP representations of
#     quantum objects.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import Optional
import warnings
import numpy as np
import qutip as qt

def convert_diagnostic_to_qobj(data) -> Optional[qt.Qobj]:
    """
    Given data deserialized from JSON diagnostics emitted by a simulator,
    attempts to convert to a QuTiP quantum object, returning the converted
    object if possible and `None` otherwise.
    """

    # Try to convert data to a Qobj if possible.
    if "amplitudes" in data:
        # Got a state vector, so convert to a Qobj with type=ket.
        # The serialization for these state vectors is defined at:
        # https://github.com/microsoft/iqsharp/blob/1015192aedababc3fe6d64e6def5838ea5eaab2f/src/Jupyter/Visualization/StateDisplayEncoders.cs#L113
        return qt.Qobj(
            [[z['Real'] + 1j * z['Imaginary'] for z in data['amplitudes']]],
            dims=[[1] * data['n_qubits'], [2] * data['n_qubits']]
        ).dag()

    # The State UDT case for density operators as represented in C# looks like:
    # {n_qubits: ..., data: {Mixed: {...}}}
    # Thus, we look for something that has both n_qubits and data, then look
    # for what the only key in data is.
    elif 'n_qubits' in data and 'data' in data:
        n_qubits = data['n_qubits']
        kind = list(data['data'].keys())[0]

        # TODO: Support kinds other than Mixed.
        if kind != 'Mixed':
            return None

        state_data = data['data'][kind]

        # Got a density operator, so convert to a Qobj with type=oper.
        # The serialization for these density operators is defined at:
        # https://github.com/microsoft/qsharp-runtime/blob/1334dc8cefb447e65feca66c463bcd77421bd5a2/src/Simulation/Simulators/OpenSystemsSimulator/DataModel/State.cs#L99
        arr = np.array(state_data['data']).reshape((2 ** n_qubits, ) * 2 + (2, ))
        qobj = qt.Qobj(
            arr[..., 0] + 1j * arr[..., 1],
            dims=[[2] * n_qubits] * 2
        )
        if not qobj.isherm or abs(qobj.tr() - 1.0) >= 1e-8:
            warnings.warn("Expected a density operator, but failed hermicity and/or trace check.")
        return qobj

    elif 'Qubits' in data and 'Data' in data:
        # Got a unitary operator, so convert to a Qobj with type=oper.
        # TODO: Find a better way of identifying unitary operators.
        #       This is very much so a hack.
        # The serialization for these unitary operators is defined at:
        # https://github.com/microsoft/QuantumLibraries/blob/687692d75af05709f0f418f8f9715cdd67c9e572/Standard/src/Diagnostics/Emulation/DataStructures.cs#L19
        n_qubits = len(data['Qubits'])
        arr = np.array(data['Data']).reshape((2 ** n_qubits, ) * 2 + (2, ))
        return qt.Qobj(
            arr[..., 0] + 1j * arr[..., 1],
            dims=[[2] * n_qubits] * 2
        )

    print(f"Got unexpected diagnostic {data}.")

    return None



#!/bin/env python
# -*- coding: utf-8 -*-
##
# loader.py: Support for exposing Q# namespaces as Python modules.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import sys
from types import ModuleType, new_class
import importlib
from importlib.abc import MetaPathFinder, Loader
import qsharp

from typing import Iterable, List, Optional, Any, Dict, Tuple

import logging
logger = logging.getLogger(__name__)

class QSharpModuleFinder(MetaPathFinder):
    def find_module(self, full_name : str, path : Optional[str] = None) -> Loader:
        # We expose Q# namespaces as their own root-level packages.
        # E.g.:
        #     >>> import Microsoft.Quantum.Intrinsic as mqi
        # Thus, we need to check if the full name is one that that we can
        # sensibly load before we proceed.

        # To check the full name, we ask the client rather than going through
        # the public API for the qsharp package, so that we can check if the
        # client is currently busy. This can happen if anything below us in
        # meta_path needs to handle an import during an execute; this is the
        # case when ZeroMQ needs to import additional functionality from a
        # Cython module to handle a message.
        # See https://github.com/Microsoft/QuantumLibraries/issues/69 for an
        # example of this failure modality.

        # If the client is busy, we'll want to forego this request to find a
        # module and return None early.
        if qsharp.client.busy:
            return None

        # At this point, we should be safe to rely on the public API again.
        ops = qsharp.get_available_operations_by_namespace()

        if full_name not in ops:
            # We may have been given part of the qualified name of a namespace.
            # E.g., if we try to import Microsoft.Quantum.Intrinsic, we'll
            # see calls with "Microsoft" and "Microsoft.Quantum" first.
            if not any(
                ns_name.startswith(full_name + ".")
                for ns_name in ops
            ):
                return None

        return QSharpModuleLoader()

class QSharpModuleLoader(Loader):
    def load_module(self, full_name : str):
        logger.debug(f"Trying to load {full_name} as a Q# namespace.")
        if full_name in sys.modules:
            return sys.modules[full_name]

        module = QSharpModule(full_name, full_name, self)

        # Register the new module.
        sys.modules.setdefault(full_name, module)
        return module

class QSharpCallable(object):
    _name : str
    def __init__(self, callable_name : str, source : str):
        self._name = callable_name
        self.source = source

    def __repr__(self) -> str:
        return f"<Q# callable {self._name}>"

    def __call__(self, **kwargs) -> Any:
        """
        Executes this function or operation on the QuantumSimulator target
        machine, returning its output as a Python object.
        """
        return self.simulate(**kwargs)

    def simulate(self, **kwargs) -> Any:
        """
        Executes this function or operation on the QuantumSimulator target
        machine, returning its output as a Python object.
        """
        return qsharp.client.simulate(self, **kwargs)

    def toffoli_simulate(self, **kwargs) -> Any:
        """
        Executes this function or operation on the ToffoliSimulator target
        machine, returning its output as a Python object.
        """
        return qsharp.client.toffoli_simulate(self, **kwargs)

    def estimate_resources(self, **kwargs) -> Dict[str, int]:
        return qsharp.client.estimate(self, **kwargs)

    def trace(self, **kwargs) -> Any:
        """
        Returns a structure representing the set of gates and qubits
        used to execute this operation.
        """
        return qsharp.client.trace(self, **kwargs)


class QSharpModule(ModuleType):
    _qs_name : str

    def __init__(self, full_name : str, qs_name : str, loader : QSharpModuleLoader):
        super().__init__(full_name)
        self._qs_name = qs_name
        self.__file__ = f"qsharp:{qs_name}"
        self.__path__ = []
        self.__loader__ = loader

    def _all_sub_namespaces_as_parts(self) -> Iterable[Tuple[str]]:
        qs_namespaces = qsharp.get_available_operations_by_namespace().keys()
        all_namespaces = set()
        for ns in qs_namespaces:
            parts = tuple(ns.split("."))
            all_namespaces.add(parts)
            for idx_part in range(1, len(parts)):
                all_namespaces.add(tuple(parts[:-idx_part]))

        return all_namespaces

    def _immediate_sub_namespaces(self) -> Iterable[str]:
        parts = tuple(self._qs_name.split("."))
        for ns in self._all_sub_namespaces_as_parts():
            if ns[:len(parts)] == parts and len(ns) > len(parts):
                yield ns[len(parts)]

    def _all_sub_namespaces(self) -> Iterable[str]:
        return [
            ".".join(ns)
            for ns in self._all_sub_namespaces_as_parts()
            if ".".join(ns).startswith(self._qs_name + ".")
        ]

    def __dir__(self) -> Iterable[str]:
        ops = qsharp.get_available_operations_by_namespace()
        return list(sorted(
            list(self._immediate_sub_namespaces()) + 
            ops.get(self._qs_name, [])
        ))

    def __getattr__(self, name):
        ops = qsharp.get_available_operations_by_namespace()
        # NB: Our Q# namespace name may not exist as a key, as the namespace
        #     name may be a prefix (e.g.: `Microsoft` and `Microsoft.Quantum.`
        #     may be empty, even though `Microsoft.Quantum.Intrinsic` is not).
        #
        #     Thus, we need to look for sub-namespaces as well as callables.
        #     While subnamespaces aren't a concept in Q#, they are in Python,
        #     and are needed to make tab completion on imports work correctly.
        #
        #     Start by looking for sub-namespaces.
        sub_namespaces = list(self._all_sub_namespaces())
        qualified_name = f"{self._qs_name}.{name}"
        if qualified_name in sub_namespaces:
            return self.__loader__.load_module(qualified_name)

        if self._qs_name in ops and name in ops[self._qs_name]:
            op_cls = new_class(name, (QSharpCallable, ))

            # Copy over metadata from the operation's header.
            metadata = qsharp.client.get_operation_metadata(f"{self._qs_name}.{name}")
            op_cls.__doc__ = metadata.get('documentation', '')
            op_cls.__file__ = metadata.get('source', None)
            return op_cls(f"{self._qs_name}.{name}", "workspace")
        raise AttributeError(f"Q# namespace {self._qs_name} does not contain a callable {name}.")

    def __repr__(self) -> str:
        return f"<module '{self._qs_name}' (Q# namespace)>"



#!/bin/env python
# -*- coding: utf-8 -*-
##
# iqsharp.py: Client for the IQ# Jupyter kernel.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""
This module allow for using experimental features of the Quantum Development Kit,
including noisy simulators for Q# programs.
"""

## DESIGN NOTES ##

# The functions in this module may take dependencies on QuTiP and NumPy,
# while neither of those are dependencies of the qsharp package on the whole.
# To avoid those becoming hard dependencies, we do not import either package
# at the top of this module, but do so inside each function that uses those
# dependencies.
#
# This has some performance and code maintenance implications, but allows for
# being permissive with respect to dependencies.

## IMPORTS ##

from typing import Any, FrozenSet, List, Tuple
import qsharp
from qsharp.loader import QSharpCallable
import json
import dataclasses

from typing import Any, Union

from qsharp.types import Pauli

## EXPORTS ##

__all__ = [
    "enable_noisy_simulation",
    "get_noise_model",
    "set_noise_model",
    "get_noise_model_by_name",
    "set_noise_model_by_name",

    # SequenceProcess process data model
    "SequenceProcess",

    # Mixed pauli data model
    "MixedPauli",

    # CHP decomposition data model
    "ChpDecompositionProcess",
    "Hadamard",
    "Cnot",
    "Phase",
    "AdjointPhase"
]

## PUBLIC FUNCTIONS ##

def enable_noisy_simulation():
    """
    Enables the `.simulate_noise` method to be used on Python objects
    representing Q# operations, allowing for Q# programs to be simulated using
    experimental simulators.

    Noisy simulation is controlled by the :func:`~qsharp.experimental.get_noise_model`,
    :func:`~qsharp.experimental.set_noise_model` and :func:`~qsharp.experimental.set_noise_model_by_name`
    functions, and by the `opensim.nQubits` and `opensim.representation` keys
    of the :any:`qsharp.config` object.
    """
    # Prevent Python from thinking that qsharp is a local variable.
    import qsharp

    # Try to import optional packages used by noise modeling.
    optional_dependencies = []
    try:
        import numpy as np
        optional_dependencies.append(f"numpy:{np.__version__}")
    except:
        np = None

    try:
        import qutip as qt
        optional_dependencies.append(f"qutip:{qt.__version__}")
    except:
        qt = None

    # Tell the kernel to turn on experimental features.
    try:
        content = {
            "feature_name": "noisy_simulation",
            "optional_dependencies": optional_dependencies
        }
        msg = qsharp.client.kernel_client.session.msg('iqsharp_python_enable_experimental', content)
        qsharp.client.kernel_client.shell_channel.send(msg)
    except:
        pass

    # Actually attach the new method to the type used for exposing Q# callables
    # to Python.
    def simulate_noise(self, **kwargs):
        return qsharp.client._simulate_noise(self, **kwargs)

    QSharpCallable.simulate_noise = simulate_noise

    # Register the experimental feature with qsharp.component_versions so that
    # it shows up in notebooks nicely.
    version = {'simulators': qsharp.client._execute('%experimental.build_info')}
    if qsharp._experimental_versions is not None:
        qsharp._experimental_versions.update(version)
    else:
        qsharp._experimental_versions = version

    # Finally, if we're in IPython, expose experimental magic commands to
    # the notebook interface.
    try:
        if __IPYTHON__:
            import qsharp.ipython_magic
            qsharp.ipython_magic.register_experimental_magics()
    except NameError:
        pass

def get_noise_model():
    """
    Returns the current noise model used in simulating Q# programs with the
    `.simulate_noise` method.
    """
    noise_model = convert_to_arrays(qsharp.client._get_noise_model())
    # Convert {"Mixed": ...} and so forth to qobj.
    return convert_to_qobjs(noise_model)

def get_noise_model_by_name(name: str):
    """
    Returns the built-in noise model with a given name.

    :param name: The name of the noise model to be returned (either `ideal`
        or `ideal_stabilizer`).
    """
    noise_model = convert_to_arrays(qsharp.client._get_noise_model_by_name(name))
    # Convert {"Mixed": ...} and so forth to qobj.
    return convert_to_qobjs(noise_model)

def set_noise_model(noise_model):
    """
    Sets the current noise model used in simulating Q# programs with the
    `.simulate_noise` method.
    """
    json_data = dumps(convert_to_rust_style(noise_model))
    qsharp.client._set_noise_model(json_data)

def set_noise_model_by_name(name):
    """
    Sets the current noise model used in simulating Q# programs with the
    `.simulate_noise` method to a built-in noise model, given by name.

    :param name: The name of the noise model to be returned (either `ideal`
        or `ideal_stabilizer`).
    """
    qsharp.client._set_noise_model_by_name(name)

## PUBLIC DATA MODEL ##

@dataclasses.dataclass
class SequenceProcess():
    n_qubits: int
    processes: List[Any]

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                "Sequence": list(map(_as_jobj, self.processes))
            }
        }

@dataclasses.dataclass
class Hadamard():
    idx_target: int

    def _as_jobj(self):
        return {
            'Hadamard': self.idx_target
        }

@dataclasses.dataclass
class Phase():
    idx_target: int

    def _as_jobj(self):
        return {
            'Phase': self.idx_target
        }

@dataclasses.dataclass
class AdjointPhase():
    idx_target: int

    def _as_jobj(self):
        return {
            'AdjointPhase': self.idx_target
        }

@dataclasses.dataclass
class Cnot():
    idx_control: int
    idx_target: int

    def _as_jobj(self):
        return {
            'Cnot': [self.idx_control, self.idx_target]
        }

@dataclasses.dataclass
class ChpDecompositionProcess():
    n_qubits: int
    operations: List[Union[Hadamard, Cnot, Phase, AdjointPhase]]

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'ChpDecomposition': list(map(_as_jobj, self.operations))
            }
        }

@dataclasses.dataclass
class MixedPauliProcess():
    n_qubits: int
    operators: List[Tuple[float, Union[List[qsharp.Pauli]], str]]

    def _as_jobj(self):
        return {
            'n_qubits': self.n_qubits,
            'data': {
                'MixedPauli': [
                    [
                        pr,
                        [
                            (qsharp.Pauli[p] if isinstance(p, str) else qsharp.Pauli(p)).name
                            for p in ops
                        ]
                    ]
                    for (pr, ops) in self.operators
                ]
            }
        }

## PRIVATE FUNCTIONS ##

literal_keys = frozenset(['ChpDecomposition'])

def _as_jobj(o, default=lambda x: x):
    import numpy as np
    if isinstance(o, np.ndarray):
        # Use Rust-style arrays.
        return {
            'v': 1,
            'dim': list(o.shape),
            'data': o.reshape((-1, )).tolist()
        }
    elif hasattr(o, '_as_jobj'):
        return o._as_jobj()

    return default(o)

class NoiseModelEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return _as_jobj(o, super().default)

def dumps(obj: Any) -> str:
    """
    Wraps json.dumps with a custom JSONEncoder class to cover types used in
    noise model serialization.
    """
    return json.dumps(
        obj=obj,
        cls=NoiseModelEncoder
    )

def is_rust_style_array(json_obj):
    return (
        isinstance(json_obj, dict) and
        'v' in json_obj and
        'data' in json_obj and
        'dim' in json_obj and
        json_obj['v'] == 1
    )

def rust_style_array_as_array(json_obj, as_complex : bool = True):
    import numpy as np
    dims = json_obj['dim'] + [2] if as_complex else json_obj['dim']
    arr = np.array(json_obj['data']).reshape(dims)
    if as_complex:
        arr = arr.astype(complex)
        return arr[..., 0] + 1j * arr[..., 1]
    else:
        return arr

def convert_to_arrays(json_obj, as_complex : bool = True):
    return (
        # Add on a trailing index of length 2.
        rust_style_array_as_array(json_obj, as_complex=as_complex)
        if is_rust_style_array(json_obj) else
        {
            key:
                value
                if key in literal_keys else

                convert_to_arrays(value, as_complex=key != 'table')
                if isinstance(value, dict) else

                [convert_to_arrays(element) for element in value]
                if isinstance(value, list) else

                value

            for key, value in json_obj.items()
        }
    )

def arr_to_qobj(arr):
    import qutip as qt
    import numpy as np
    return qt.Qobj(arr, dims=[[2] * int(np.log2(arr.shape[1]))] * 2)

def convert_to_qobjs(json_obj):
    import qutip as qt
    return (
        arr_to_qobj(json_obj['data']['Mixed'])
        if 'data' in json_obj and 'Mixed' in json_obj['data'] else

        arr_to_qobj(json_obj['data']['Unitary'])
        if 'data' in json_obj and 'Unitary' in json_obj['data'] else

        qt.kraus_to_super([arr_to_qobj(op) for op in json_obj['data']['KrausDecomposition']])
        if 'data' in json_obj and 'KrausDecomposition' in json_obj['data'] else

        {
            key:
                value
                if key in literal_keys else

                # Recurse if needed...
                convert_to_qobjs(value)
                if isinstance(value, dict) else

                [convert_to_qobjs(element) for element in value]
                if isinstance(value, list) else

                # Just return value if there's nothing else to do.
                value
            for key, value in json_obj.items()
        }
    )

def arr_to_rust_style(arr):
    import numpy as np
    return {
        'v': 1,
        'dim': list(arr.shape),
        'data': np.moveaxis(np.array([arr.real, arr.imag]), 0, -1).reshape((-1, 2)).tolist()
    }

def qobj_to_rust_style(qobj, expect_state=False):
    import qutip as qt
    import numpy as np
    data = None
    n_qubits = 1
    # Figure out what kind of qobj we have and convert accordingly.
    if qobj.type == 'oper':
        n_qubits = len(qobj.dims[0])
        data = {
            'Mixed' if expect_state else 'Unitary':
            arr_to_rust_style(qobj.data.todense())
        }
    elif qobj.type == 'super':
        n_qubits = len(qobj.dims[0][0])
        data = {
            'KrausDecomposition': arr_to_rust_style(
                np.array([
                    op.data.todense()
                    for op in qt.to_kraus(qobj)
                ])
            )
        }
    return {
        "n_qubits": n_qubits,
        "data": data
    }

def convert_to_rust_style(json_obj, expect_state=False):
    import qutip as qt
    return (
        qobj_to_rust_style(json_obj, expect_state=expect_state)
        if isinstance(json_obj, qt.Qobj) else

        list(map(convert_to_rust_style, json_obj))
        if isinstance(json_obj, list) else

        {
            key: convert_to_rust_style(value, expect_state=key == 'initial_state')
            for key, value in json_obj.items()
        }
        if isinstance(json_obj, dict) else

        json_obj
    )



#!/bin/env python
# -*- coding: utf-8 -*-
##
# __init__.py: Root module for the qsharp package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

"""

"""

## IMPORTS ##

import sys
from contextlib import contextmanager
from typing import Any, List, Dict, Union
from collections import defaultdict
from distutils.version import LooseVersion

from qsharp.clients import _start_client
from qsharp.clients.iqsharp import IQSharpError
from qsharp.loader import QSharpCallable, QSharpModuleFinder
from qsharp.config import Config
from qsharp.packages import Packages
from qsharp.projects import Projects
from qsharp.types import Result, Pauli
try:
    from qsharp.version import __version__
except:
    __version__ = "<unknown>"

## EXPORTS ##

__all__ = [
    'compile', 'reload',
    'get_available_operations', 'get_available_operations_by_namespace',
    'get_workspace_operations',
    'config',
    'packages',
    'projects',
    'IQSharpError',
    'Result', 'Pauli'
]

## FUNCTIONS ##

def compile(code : str) -> Union[None, QSharpCallable, List[QSharpCallable]]:
    """
    Given a string containing Q# source code, compiles it into the current
    workspace and returns one or more Q# callable objects that can be used to
    invoke the new code.

    :param code: A string containing Q# source code to be compiled.
    :returns: A list of callables compiled from `code`, or a callable if exactly
        one callable is found.
    """
    compiled = client.compile(code)
    if compiled is None:
        return None

    ops = [
        QSharpCallable(op, "snippets")
        for op in client.compile(code)
    ]
    if len(ops) == 1:
        return ops[0]
    else:
        return ops


def reload() -> None:
    """
    Reloads the current IQ# workspace, recompiling source files in the
    workspace.

    If the workspace fails to compile (e.g., because of a missing package),
    Q# compilation errors are raised as an exception.
    """
    client.reload()

def get_available_operations() -> List[str]:
    """
    Returns a list containing the names of all operations and functions defined
    in the current workspace, and that have been compiled dynamically from
    snippets.
    """
    return client.get_available_operations()

def get_workspace_operations() -> List[str]:
    """
    Returns a list containing the names of all operations and functions defined
    in the current workspace, excluding dynamically compiled snippets.
    """
    return client.get_workspace_operations()

def get_available_operations_by_namespace() -> Dict[str, List[str]]:
    """
    Returns a dictionary from namespaces to lists of operations and functions
    defined in the current workspace that are members of each namespace.
    """
    ops = get_available_operations()
    by_ns = defaultdict(list)

    for qualified_name in ops:
        idx_last_dot = qualified_name.rfind(".")
        ns_name = qualified_name[:idx_last_dot]
        op_name = qualified_name[idx_last_dot + 1:]

        by_ns[ns_name].append(op_name)

    return dict(by_ns.items())

def component_versions() -> Dict[str, LooseVersion]:
    """
    Returns a dictionary from components of the IQ# kernel to their
    versions.
    """
    versions = client.component_versions()
    # Add in the qsharp Python package itself.
    versions["qsharp"] = LooseVersion(__version__)
    # If any experimental features are enabled, report them here.
    if _experimental_versions is not None:
        versions['experimental'] = _experimental_versions
    return versions

@contextmanager
def capture_diagnostics(passthrough: bool = False, as_qobj: bool = False) -> List[Any]:
    """
    Returns a context manager that captures diagnostics output from running Q#
    programs into a list.

    For example, to capture `DumpMachine` calls from a Q# operation run on the
    full-state simulator:

    .. code-block:: qsharp

        namespace Sample {
            open Microsoft.Quantum.Intrinsic;
            open Microsoft.Quantum.Diagnostics;

            operation RunMain() : Unit {
                use q = Qubit();
                within {
                    H(q);
                } apply {
                    DumpMachine();
                }
            }
        }

    .. code-block:: python

        import qsharp
        from Sample import RunMain

        with qsharp.capture_diagnostics() as diagnostics:
            RunMain.simulate()

        print(len(diagnostics)) # will print 1

    :param passthrough: If `True`, captured diagnostics will also be displayed
        as normal. By default, diagnostic outputs captured by this context
        manager will not be displayed.
    :param as_qobj: If `True`, this context manager will attempt to convert
        captured diagnostics representing quantum states and operations into
        QuTiP objects. This option requires that QuTiP is installed and
        can be imported.
    """
    # Before proceeding, check that if we were asked to convert to qobj data
    # that we can actually import qutip.
    if as_qobj:
        try:
            # We don't actually need QuTiP here, but are only importing to capture
            # exceptions as early as possible so as to provide actionable error
            # messages to the user.
            import qutip as _
        except ImportError as ex:
            raise ImportError("as_qobj was set to `True`, but cannot convert captured diagnostics to QObj since QuTiP failed to import.") from ex

        from qsharp.qobj import convert_diagnostic_to_qobj

    processed_data = []
    with client.capture_diagnostics(passthrough=passthrough) as data:
        yield processed_data

        # Apply any postprocessing needed here and append to processed_data.
        for diagnostic in data:
            if as_qobj:
                converted = convert_diagnostic_to_qobj(diagnostic)
                if converted is not None:
                    diagnostic = converted
            processed_data.append(diagnostic)

## STARTUP ##

client = _start_client()
config = Config(client)
packages = Packages(client)
projects = Projects(client)
_experimental_versions = None

# Make sure that we're last on the meta_path so that actual modules are loaded
# first.
sys.meta_path.append(QSharpModuleFinder())

# If using IPython, forward some useful IQ# magic commands as IPython magic
# commands and define a couple new magic commands for IPython.
try:
    if __IPYTHON__:
        import qsharp.ipython_magic
        qsharp.ipython_magic.register_magics()
except NameError:
    pass

# Needed to recognize PEP 420 packages as subpackages.
import pkg_resources
pkg_resources.declare_namespace(__name__)


#!/bin/env python
# -*- coding: utf-8 -*-
##
# setup.py: Installs Python host functionality for Q#.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

## IMPORTS ##

import setuptools
import os

## VERSION INFORMATION ##
# Our build process sets the PYTHON_VERSION environment variable to a version
# string that is compatible with PEP 440, and so we inherit that version number
# here and propagate that to qsharp/version.py.
#
# To make sure that local builds still work without the same environment
# variables, we'll default to 0.0.0.1 as a development version.

version = os.environ.get('PYTHON_VERSION', '0.0.0.1')

with open('./qsharp/version.py', 'w') as f:
    f.write(f'''# Auto-generated file, do not edit.
##
# version.py: Specifies the version of the qsharp package.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
__version__ = "{version}"
is_conda = False
_user_agent_extra = "[{version}]"
''')

## DESCRIPTION ##
# The long description metadata passed to setuptools is used to populate the
# PyPI page for this package. Thus, we'll generate the description by using the
# same README.md file that we use in the GitHub repo.

with open("./README.md", "r") as fh:
    long_description = fh.read()

## SETUPTOOLS INVOCATION ##

setuptools.setup(
    name="qsharp-core",
    version=version,
    author="Microsoft",
    author_email="que-contacts@microsoft.com",
    description="Python client for Q#, a domain-specific quantum programming language",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/iqsharp",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'jupyter_client',
        'pyzmq<20.0.0'     # due to incompatibility of IQ# with pyzmq>=20.0.0
    ]
)



#!/bin/env python
# -*- coding: utf-8 -*-
##
# types.py: Provides types for interoperability with Q#.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from enum import IntEnum
import random

try:
    import qutip as qt
except ImportError:
    qt = None

## ENUMS ######################################################################

class Result(IntEnum):
    Zero = 0
    One = 1

    @classmethod
    def sample(cls):
        return random.choice(list(cls))

class Pauli(IntEnum):
    I = 0b00
    X = 0b01
    Y = 0b11
    Z = 0b10

    @classmethod
    def sample(cls):
        return random.choice(list(cls))

    if qt is not None:
        def as_qobj(self):
            if self == Pauli.I:
                return qt.qeye(2)
            elif self == Pauli.X:
                return qt.sigmax()
            elif self == Pauli.Y:
                return qt.sigmay()
            elif self == Pauli.Z:
                return qt.sigmaz()
            else:
                raise ValueError(f"Unrecognized Pauli value {self}.")



#!/bin/env python
# -*- coding: utf-8 -*-
##
# test_iqsharp.py: Tests basic Q#/Python interop functionality.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

## IMPORTS ##

import json
import numpy as np
import os
import pytest
import qsharp
import qsharp.experimental
qsharp.experimental.enable_noisy_simulation()
from .utils import set_environment_variables

print ( qsharp.component_versions() )

## SETUP ##

@pytest.fixture(scope="session", autouse=True)
def session_setup():
    set_environment_variables()

try:
    import qutip as qt
except ImportError:
    qt = None

skip_if_no_qutip = pytest.mark.skipif(qt is None, reason="Test requires QuTiP.")

# If we can't import Microsoft as a Python package, then that means that
# the Q# workspace isn't present — probably because the qsharp-core package
# has been installed, since `Operations.qs` isn't part of the package data.
try:
    import Microsoft.Quantum.SanityTests
    workspace_present = True
except ImportError:
    workspace_present = False

skip_if_no_workspace = pytest.mark.skipif(not workspace_present, reason="Local Q# workspace not available in installed package.")

# Finally, not all tests work well from within conda build environments, so
# we disable those tests here.
is_conda = getattr(qsharp.__version__, "is_conda", False) or os.environ.get("QSHARP_PY_ISCONDA", "False").lower() == "true"
skip_if_conda = pytest.mark.skipif(is_conda, reason="Test is not supported from conda-build.")

## TESTS ##

@skip_if_no_workspace
def test_simulate():
    """
    Checks that a simple simulate works correctly, both using callable() and 
    callable.simulate()
    """
    from Microsoft.Quantum.SanityTests import HelloQ, HelloAgain
    assert HelloQ() == HelloQ.simulate() == ()
    assert HelloAgain(
        count=1, name="Ada") == HelloAgain.simulate(count=1, name="Ada")


def test_toffoli_simulate():
    foo = qsharp.compile("""
        open Microsoft.Quantum.Measurement;

        operation Foo() : Result {
            using (q = Qubit()) {
                X(q);
                return MResetZ(q);
            }
        }
    """)
    assert foo.toffoli_simulate() == 1

@skip_if_no_workspace
def test_tuples():
    """
    Checks that tuples are correctly encoded both ways.
    """
    from Microsoft.Quantum.SanityTests import IndexIntoTuple
    r = IndexIntoTuple.simulate(count=2, tuples=[(0, "Zero"), (1, "One"), (0, "Two"), (0, "Three")])
    assert r == (0, "Two")

@skip_if_no_workspace
def test_numpy_types():
    """
    Checks that numpy types are correctly encoded.
    """
    from Microsoft.Quantum.SanityTests import IndexIntoNestedArray, IndexIntoTuple

    r = IndexIntoNestedArray.simulate(index1=1, index2=0, nestedArray=np.array([[100, 101], np.array([110, 111], dtype=np.int64)]))
    assert r == 110
    
    tuples = [(0, "Zero"), (1, "One")]
    tuples_array = np.empty(len(tuples), dtype=object)
    tuples_array[:] = tuples
    r = IndexIntoTuple.simulate(count=1, tuples=tuples_array)
    assert r == (1, "One")

@skip_if_no_workspace
def test_result():
    """
    Checks that Result-type arguments are handled correctly.
    """
    from Microsoft.Quantum.SanityTests import EchoResult
    r = EchoResult.simulate(input=qsharp.Result.One)
    assert r == qsharp.Result.One    

    r = EchoResult.simulate(input=1)
    assert r == qsharp.Result.One

    # Current behavior is that non-integer values will get rounded to
    # the nearest integer then converted to a Result. Once that behavior
    # is fixed, this test should be updated to ensure that the following
    # code throws a qsharp.IQSharpError exception.
    # See https://github.com/microsoft/qsharp-runtime/issues/376.
    r = EchoResult.simulate(input=0.2)
    assert r == qsharp.Result.Zero

@skip_if_no_workspace
def test_long_tuple():
    """
    Checks that a 10-tuple argument and return value are handled correctly.
    """
    ten_tuple = (0, 10, 20, 30, 40, 50, 60, 70, 80, 90)
    
    from Microsoft.Quantum.SanityTests import IndexIntoTenTuple
    r = IndexIntoTenTuple.simulate(index=3, tenTuple=ten_tuple)
    assert r == 30
    r = IndexIntoTenTuple.simulate(index=8, tenTuple=ten_tuple)
    assert r == 80
    
    from Microsoft.Quantum.SanityTests import EchoTenTuple
    r = EchoTenTuple.simulate(tenTuple=ten_tuple)
    assert r == ten_tuple

@skip_if_no_workspace
def test_paulis():
    """
    Checks that Pauli-type and arrays of Pauli-type arguments are handled correctly.
    """
    from Microsoft.Quantum.SanityTests import SwapFirstPauli

    paulis = [qsharp.Pauli.Z, qsharp.Pauli.Z, qsharp.Pauli.Z]
    pauliToSwap = qsharp.Pauli.X
    r = SwapFirstPauli.simulate(paulis=paulis, pauliToSwap=pauliToSwap)

    assert r[0] == [qsharp.Pauli.X, qsharp.Pauli.Z, qsharp.Pauli.Z]
    assert r[1] == qsharp.Pauli.Z

    # Should also work with string representation
    paulis = ["PauliZ", "PauliZ", "PauliZ"]
    pauliToSwap = "PauliX"
    r = SwapFirstPauli.simulate(paulis=paulis, pauliToSwap=pauliToSwap)

    assert r[0] == [qsharp.Pauli.X, qsharp.Pauli.Z, qsharp.Pauli.Z]
    assert r[1] == qsharp.Pauli.Z

@skip_if_no_workspace
def test_estimate():
    """
    Verifies that resource estimation works.
    """
    from Microsoft.Quantum.SanityTests import HelloAgain
    r = HelloAgain.estimate_resources(count=4, name="estimate test")
    assert r['Measure'] == 8
    assert r['QubitClifford'] == 1
    assert r['BorrowedWidth'] == 0

@skip_if_no_workspace
def test_trace():
    """
    Verifies the trace commands works.
    """
    from Microsoft.Quantum.SanityTests import HelloAgain
    r = HelloAgain.trace(count=1, name="trace test")
    print(r)
    assert len(r['qubits']) == 1
    assert len(r['operations']) == 1
    assert len(r['operations'][0]['children']) == 2
    assert r['operations'][0]['gate'] == 'HelloAgain'
    assert r['operations'][0]['children'][0]['gate'] == 'M'
    assert r['operations'][0]['children'][1]['gate'] == 'Reset'


def test_simple_compile():
    """
    Verifies that compile works
    """
    op = qsharp.compile( """
    operation HelloQ() : Result
    {
        Message($"Hello from quantum world!"); 
        return Zero;
    }
""")
    r = op.simulate()
    assert r == 0
    

def test_multi_compile():
    """
    Verifies that compile works and that operations
    are returned in the correct order
    """
    ops = qsharp.compile( """
    operation HelloQ() : Result
    {
        Message($"Hello from quantum world!"); 
        return One;
    }

    operation Hello2() : Result
    {
        Message($"Will call hello."); 
        return HelloQ();
    }
""")
    assert "HelloQ" == ops[0]._name
    assert "Hello2" == ops[1]._name

    r = ops[1].simulate()
    assert r == qsharp.Result.One


def test_config():
    """
    Verifies get and set of config settings of various types
    """
    qsharp.config["dump.basisStateLabelingConvention"] = "Bitstring"
    qsharp.config["dump.truncateSmallAmplitudes"] = True
    qsharp.config["dump.truncationThreshold"] = 1e-6

    assert qsharp.config["dump.basisStateLabelingConvention"] == "Bitstring"
    assert qsharp.config["dump.truncateSmallAmplitudes"] == True
    assert qsharp.config["dump.truncationThreshold"] == 1e-6

@skip_if_conda
def test_packages():
    """
    Verifies default package command
    """
    pkg_count = len(qsharp.packages._client.get_packages())

    with pytest.raises(Exception):
        qsharp.packages.add('Invalid.Package.!!!!!!')
    assert pkg_count == len(qsharp.packages._client.get_packages())

    # We test using a non-QDK package name to avoid possible version conflicts.
    qsharp.packages.add('Microsoft.Extensions.Logging')
    assert (pkg_count+1) == len(qsharp.packages._client.get_packages())


def test_projects(tmp_path):
    """
    Verifies default project command
    """
    assert 0 == len(qsharp.projects._client.get_projects())

    with pytest.raises(Exception):
        qsharp.projects.add('../InvalidPath/InvalidProject.txt')
    assert 0 == len(qsharp.projects._client.get_projects())

    temp_project_path = tmp_path / 'temp_iqsharp_pytest_project.csproj'
    temp_project_path.write_text(f'''
        <Project Sdk="Microsoft.Quantum.Sdk/0.12.20072031">
            <PropertyGroup>
                <TargetFramework>netstandard2.1</TargetFramework>
                <IncludeQsharpCorePackages>false</IncludeQsharpCorePackages>
            </PropertyGroup>
        </Project>
    ''')

    qsharp.projects.add(str(temp_project_path))
    assert 1 == len(qsharp.projects._client.get_projects())

class TestCaptureDiagnostics:
    def test_basic_capture(self):
        dump_plus = qsharp.compile("""
            open Microsoft.Quantum.Diagnostics;

            operation DumpPlus() : Unit {
                use qs = Qubit[2];
                within {
                    H(qs[0]);
                    H(qs[1]);
                } apply {
                    DumpMachine();
                }
            }
        """)

        with qsharp.capture_diagnostics(as_qobj=False) as captured:
            dump_plus.simulate()

        assert 1 == len(captured)
        expected = """
            {
                "diagnostic_kind": "state-vector",
                "qubit_ids": [0, 1],
                "n_qubits": 2,
                "amplitudes": [{"Real": 0.5000000000000001, "Imaginary": 0.0, "Magnitude": 0.5000000000000001, "Phase": 0.0}, {"Real": 0.5000000000000001, "Imaginary": 0.0, "Magnitude": 0.5000000000000001, "Phase": 0.0}, {"Real": 0.5000000000000001, "Imaginary": 0.0, "Magnitude": 0.5000000000000001, "Phase": 0.0}, {"Real": 0.5000000000000001, "Imaginary": 0.0, "Magnitude": 0.5000000000000001, "Phase": 0.0}]
            }
        """
        del captured[0]['div_id']
        assert json.dumps(json.loads(expected)) == json.dumps(captured[0])


    @skip_if_no_qutip
    def test_capture_diagnostics_as_qobj(self):
        dump_plus = qsharp.compile("""
            open Microsoft.Quantum.Diagnostics;

            operation DumpPlus() : Unit {
                use qs = Qubit[2];
                within {
                    H(qs[0]);
                    H(qs[1]);
                } apply {
                    DumpMachine();
                }
            }
        """)

        with qsharp.capture_diagnostics(as_qobj=True) as captured:
            dump_plus.simulate()

        assert 1 == len(captured)
        assert abs(1.0 - captured[0].norm()) <= 1e-8

        import qutip as qt
        expected = qt.Qobj([[1], [1], [1], [1]], dims=[[2, 2], [1, 1]]).unit()
        assert (expected - captured[0]).norm() <= 1e-8


    @skip_if_no_qutip
    def test_capture_experimental_diagnostics_as_qobj(self):
        dump_plus = qsharp.compile("""
            open Microsoft.Quantum.Diagnostics;

            operation DumpPlus() : Unit {
                use qs = Qubit[2];
                within {
                    H(qs[0]);
                    H(qs[1]);
                } apply {
                    DumpMachine();
                }
            }
        """)

        qsharp.experimental.set_noise_model_by_name('ideal')
        qsharp.config['experimental.simulators.nQubits'] = 2
        qsharp.config['experimental.simulators.representation'] = 'mixed'

        with qsharp.capture_diagnostics(as_qobj=True) as captured:
            dump_plus.simulate_noise()

        assert 1 == len(captured)
        assert abs(1.0 - captured[0].tr()) <= 1e-8

        import qutip as qt
        expected = qt.Qobj([[1], [1], [1], [1]], dims=[[2, 2], [1, 1]]).unit()
        expected = expected * expected.dag()
        assert (expected - captured[0]).norm() <= 1e-8
